/* 
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997, 1998 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 */

#include "wconfig.h"

#ifdef BALLOON_TEXT

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#ifdef SHAPED_BALLOON
#include <X11/extensions/shape.h>
#endif

#include <stdlib.h>
#include <string.h>

#include <wraster.h>

#include "WindowMaker.h"
#include "screen.h"
#include "texture.h"
#include "wcore.h"
#include "framewin.h"
#include "icon.h"
#include "appicon.h"
#include "funcs.h"
#include "workspace.h"
#include "balloon.h"



extern WPreferences wPreferences;

typedef struct _WBalloon {
    Window window;

#ifdef SHAPED_BALLOON
    GC monoGC;
#endif
    int prevType;

    Window objectWindow;
    char *text;
    int h;
    
    WMHandlerID timer;
    
    Pixmap contents;
    
    char mapped;
} WBalloon;


#define TOP	0
#define BOTTOM	1
#define LEFT	0
#define RIGHT	2

#define TLEFT	(TOP|LEFT)
#define TRIGHT 	(TOP|RIGHT)
#define BLEFT	(BOTTOM|LEFT)
#define BRIGHT	(BRIGHT|RIGHT)


#ifdef SHAPED_BALLOON

#define 	SPACE	12


static Pixmap
makePixmap(WScreen *scr, int width, int height, int side, Pixmap *mask)
{
    WBalloon *bal = scr->balloon;
    Pixmap bitmap;
    Pixmap pixmap;
    int x, y;
    int x1, y1, x2, y2;
    XGCValues gcv;

    bitmap = XCreatePixmap(dpy, scr->root_win, width+SPACE, height+SPACE, 1);

    if (!bal->monoGC) {
	bal->monoGC = XCreateGC(dpy, bitmap, 0, NULL);
    }
    
    pixmap = XCreatePixmap(dpy, scr->root_win, width+SPACE, height+SPACE,
			   scr->w_depth);

    XSetForeground(dpy, bal->monoGC, scr->black_pixel);
    XFillRectangle(dpy, bitmap, bal->monoGC, 0, 0, width+SPACE, height+SPACE);

    if (scr->window_title_texture[0]->any.color.pixel==scr->black_pixel)
	XSetForeground(dpy, scr->draw_gc, scr->white_pixel);
    else
	XSetForeground(dpy, scr->draw_gc, scr->black_pixel);
    XFillRectangle(dpy, pixmap, scr->draw_gc, 0, 0, width+SPACE, height+SPACE);

    XSetForeground(dpy, bal->monoGC, scr->white_pixel);

    if (side & BOTTOM) {
	y = 0;
	y1 = height-2;
	y2 = height+SPACE;
    } else {
	y = SPACE;
	y1 = SPACE+2;
	y2 = 0;
    }
    if (side & RIGHT) {
	x = 0;
	x1 = width-2;
	x2 = width+SPACE;
    } else {
	x = SPACE;
	x1 = SPACE+2;
	x2 = 0;
    }

    if (scr->window_title_texture[0])
	XSetForeground(dpy, scr->draw_gc,
		       scr->window_title_texture[0]->any.color.pixel);
    else
	XSetForeground(dpy, scr->draw_gc, scr->light_pixel);

    XFillRectangle(dpy, bitmap, bal->monoGC, x, y, width, height);
    XFillRectangle(dpy, pixmap, scr->draw_gc, x+1, y+1, width-2, height-2);
    
    gcv.line_width = 2;
    XChangeGC(dpy, bal->monoGC, GCLineWidth, &gcv);
    if (scr->window_title_texture[0]->any.color.pixel==scr->black_pixel)
	gcv.foreground = scr->white_pixel;
    else
	gcv.foreground = scr->black_pixel;
    XChangeGC(dpy, scr->draw_gc, GCForeground|GCLineWidth, &gcv);

    XDrawLine(dpy, bitmap, bal->monoGC, x1, y1, x2, y2);
    XDrawLine(dpy, pixmap, scr->draw_gc, x1, y1, x2, y2);

    gcv.line_width = 1;
    XChangeGC(dpy, scr->draw_gc, GCLineWidth, &gcv);
    XChangeGC(dpy, bal->monoGC, GCLineWidth, &gcv);

    *mask = bitmap;

    return pixmap;
}


static void
showText(WScreen *scr, int x, int y, int h, int w, char *text)
{
    int width;
    int height;
    Pixmap pixmap;
    Pixmap mask;
    WFont *font = scr->title_font;
    int side = 0;
    int tx, ty;
    int bx, by;

    if (scr->balloon->contents)
	XFreePixmap(dpy, scr->balloon->contents);

    width = wTextWidth(font->font, text, strlen(text))+8;
    height = font->height + 4;
    
    if (x + width + SPACE > scr->scr_width) {
	side = RIGHT;
	bx = x - (width + SPACE);
	tx = 0;
    } else {
	side = LEFT;
	bx = x + w/2;
	if (bx + width + SPACE > scr->scr_width)
	    bx = scr->scr_width - (width + SPACE);
	tx = SPACE;
    }
    if (y - (height + SPACE) < 0) {
	side |= TOP;
	by = y;
	ty = SPACE;
    } else {
	side |= BOTTOM;
	by = y - (height + SPACE);
	ty = 0;
    }
    pixmap = makePixmap(scr, width, height, side, &mask);

    XSetForeground(dpy, scr->window_title_gc, scr->window_title_pixel[0]);

    wDrawString(pixmap, font, scr->window_title_gc, tx + 4, ty + font->y + 2,
		text, strlen(text));

    XSetWindowBackgroundPixmap(dpy, scr->balloon->window, pixmap);
    scr->balloon->contents = pixmap;
    
    XResizeWindow(dpy, scr->balloon->window, width+SPACE, height+SPACE);
    XShapeCombineMask(dpy, scr->balloon->window, ShapeBounding, 0, 0, mask,
		      ShapeSet);
    XFreePixmap(dpy, mask);
    XMoveWindow(dpy, scr->balloon->window, bx, by);
    XMapRaised(dpy, scr->balloon->window);
    
    scr->balloon->mapped = 1;
}
#else /* !SHAPED_BALLOON */
static void
showText(WScreen *scr, int x, int y, int h, int w, char *text)
{
    int width;
    int height;
    Pixmap pixmap;
    WFont *font = scr->title_font;

    if (scr->balloon->contents)
	XFreePixmap(dpy, scr->balloon->contents);

    width = wTextWidth(font->font, text, strlen(text))+8;
    height = font->height + 4;

    if (x < 0)
	x = 0;
    else if (x + width > scr->scr_width-1)
	x = scr->scr_width - width;
    
    if (y - height - 2 < 0) {
	y += h;
	if (y < 0)
	    y = 0;
    } else {
	y -= height + 2;
    }
    
    if (scr->window_title_texture[0])
	XSetForeground(dpy, scr->window_title_gc,
		       scr->window_title_texture[0]->any.color.pixel);
    else
	XSetForeground(dpy, scr->window_title_gc, scr->light_pixel);

    pixmap = XCreatePixmap(dpy, scr->root_win, width, height, scr->w_depth);
    XFillRectangle(dpy, pixmap, scr->window_title_gc, 0, 0, width, height);

    XSetForeground(dpy, scr->window_title_gc, scr->window_title_pixel[0]);

    wDrawString(pixmap, font->font, scr->window_title_gc, 4, font->y+2,
		text, strlen(text));
    
    XResizeWindow(dpy, scr->balloon->window, width, height);
    XMoveWindow(dpy, scr->balloon->window, x, y);

    XSetWindowBackgroundPixmap(dpy, scr->balloon->window, pixmap);
    XClearWindow(dpy, scr->balloon->window);
    XMapRaised(dpy, scr->balloon->window);

    scr->balloon->contents = pixmap;
    
    scr->balloon->mapped = 1;
}
#endif /* !SHAPED_BALLOON */


static void
showBalloon(WScreen *scr)
{
    int x, y;
    Window foow;
    unsigned foo, w;

    scr->balloon->timer = NULL;

    if (!XGetGeometry(dpy, scr->balloon->objectWindow, &foow, &x, &y,
		      &w, &foo, &foo, &foo)) {
	scr->balloon->prevType = 0;
	return;
    }
    showText(scr, x, y, scr->balloon->h, w, scr->balloon->text);
}



static void
frameBalloon(WObjDescriptor *object)
{
    WFrameWindow *fwin = (WFrameWindow*)object->parent;
    WScreen *scr = fwin->core->screen_ptr;

    if (fwin->titlebar != object->self
	|| !fwin->flags.is_client_window_frame) {
	wBalloonHide(scr);
	return;
    }
    if (fwin->title && fwin->flags.incomplete_title) {
	scr->balloon->h = (fwin->titlebar ? fwin->titlebar->height : 0);
	scr->balloon->text = wstrdup(fwin->title);
	scr->balloon->objectWindow = fwin->core->window;
	scr->balloon->timer = WMAddTimerHandler(BALLOON_DELAY,
						(WMCallback*)showBalloon, scr);
    }
}


static void
miniwindowBalloon(WObjDescriptor *object)
{
    WIcon *icon = (WIcon*)object->parent;
    WScreen *scr = icon->core->screen_ptr;

    if (!icon->icon_name) {
	wBalloonHide(scr);
	return;
    }
    scr->balloon->h = icon->core->height;
    scr->balloon->text = wstrdup(icon->icon_name);
    scr->balloon->objectWindow = icon->core->window;
    if (scr->balloon->prevType != 0) {
	XUnmapWindow(dpy, scr->balloon->window);
	showBalloon(scr);
    } else {
	scr->balloon->timer = WMAddTimerHandler(BALLOON_DELAY,
						(WMCallback*)showBalloon, scr);
    }
}



static void
appiconBalloon(WObjDescriptor *object)
{
    WAppIcon *aicon = (WAppIcon*)object->parent;
    WScreen *scr = aicon->icon->core->screen_ptr;
    char *tmp;

    if (aicon->command && aicon->wm_class) {
	tmp = wmalloc(strlen(aicon->command)+strlen(aicon->wm_class)+8);
	sprintf(tmp, "%s (%s)", aicon->wm_class, aicon->command);
	scr->balloon->text = tmp;
    } else if (aicon->command) {
	scr->balloon->text = wstrdup(aicon->command);
    } else if (aicon->wm_class) {
	scr->balloon->text = wstrdup(aicon->wm_class);
    } else {
	wBalloonHide(scr);
	return;
    }
    scr->balloon->h = aicon->icon->core->height;

    scr->balloon->objectWindow = aicon->icon->core->window;
    if (scr->balloon->prevType != 0) {
	XUnmapWindow(dpy, scr->balloon->window);
	showBalloon(scr);
    } else {
	scr->balloon->timer = WMAddTimerHandler(BALLOON_DELAY,
						(WMCallback*)showBalloon, scr);
    }
}



void
wBalloonInitialize(WScreen *scr)
{
    WBalloon *bal;
    XSetWindowAttributes attribs;
    unsigned long vmask;
    
    bal = wmalloc(sizeof(WBalloon));
    memset(bal, 0, sizeof(WBalloon));
    
    scr->balloon = bal;

    vmask = CWSaveUnder|CWOverrideRedirect|CWColormap;
    attribs.save_under = True;
    attribs.override_redirect = True;
    attribs.colormap = scr->w_colormap;
    
    bal->window = XCreateWindow(dpy, scr->root_win, 1, 1, 10, 10, 1,
				scr->w_depth, CopyFromParent,
				scr->w_visual, vmask, &attribs);
#if 0
    /* select EnterNotify to so that the balloon will be unmapped
     * when the pointer is moved over it */
    XSelectInput(dpy, bal->window, EnterWindowMask);
#endif
}



void
wBalloonEnteredObject(WScreen *scr, WObjDescriptor *object)
{
    WBalloon *balloon = scr->balloon;
    
    if (balloon->timer)
	WMDeleteTimerHandler(balloon->timer);
    balloon->timer = NULL;

    if (scr->balloon->text)
	free(scr->balloon->text);
    scr->balloon->text = NULL;
    
    if (!object) {
	wBalloonHide(scr);
	return;
    }
    switch (object->parent_type) {
     case WCLASS_FRAME:
	if (wPreferences.window_balloon) {
	    frameBalloon(object);
	}
	break;

     case WCLASS_DOCK_ICON:
	if (object->parent != scr->clip_icon && wPreferences.appicon_balloon)
	    appiconBalloon(object);
	break;

     case WCLASS_MINIWINDOW:
	if (wPreferences.miniwin_balloon) {
	    miniwindowBalloon(object);
	}
	break;
     case WCLASS_APPICON:
	if (wPreferences.appicon_balloon)
	    appiconBalloon(object);
	break;
     default:
	wBalloonHide(scr);
	break;
    }
    scr->balloon->prevType = object->parent_type;
}



void
wBalloonHide(WScreen *scr)
{
    if (scr) {
	if (scr->balloon->mapped) {
	    XUnmapWindow(dpy, scr->balloon->window);
	    scr->balloon->mapped = 0;
	    scr->balloon->prevType = 0;
	} else if (scr->balloon->timer) {
	    WMDeleteTimerHandler(scr->balloon->timer);
	    scr->balloon->timer = NULL;
	}
    }
}
#endif
